#!/bin/sh
#
# $Id: trafd.sh,v 1.2 2003/10/27 13:21:41 bartosh Exp $
# Written by Andrew Khlebutin <ak@ernest.ru>
#
#
# if $1=='start': start of the network traffic collector
# if $1=='stop' : dump all records to /var/tmp/trafd.$iface
#		and shutdown network traffic collector ('kill')
#
PREFIX=/usr/local

if [ -r $PREFIX/etc/trafd.conf ]; then
        . $PREFIX/etc/trafd.conf
else
	/bin/echo "Can't find $PREFIX/etc/trafd.conf. Exiting ..."
	exit 1
fi

if [ ! "${ifaces}" ] ; then
	/bin/echo `/bin/date +"%b %e %H:%M:%S"` `/bin/hostname -s` trafd.sh: Not an interface for listen!
	exit 1
fi

# All ifaces except loopback
real_ifaces="`/sbin/ifconfig -lu | /usr/bin/sed s/lo0//`"

# Check existing ifaces
for iface in $ifaces; do
    if ! (echo $real_ifaces | grep "\b$iface\b" > /dev/null 2>&1)
    then
	/bin/echo "Mismatch Interfaces configuration. Check trafd.conf. Exiting ..."
        exit 1
    fi
done

case $1 in
start)
    for iface in $ifaces; do
        pid_filename=$pid_file$iface
        if [ -f $pid_filename ]; then
		rm $pid_filename
	fi
	eval cmd=\$${iface}_cmd_param 
	eval filter=\$${iface}_filter
	cmd_param="-pri ${iface} ${cmd} ${filter}"
	$PREFIX/sbin/trafd ${cmd_param} && /bin/echo " trafd (${iface}) "
    done
;;
stop)
    for iface in $ifaces; do
	pid_filename=$pid_file$iface
	if [ -f $pid_filename ]; then
		PID="`cat $pid_filename`"
		/bin/kill -QUIT ${PID}
	else
	    /bin/echo "`/bin/date +"%b %e %H:%M:%S"` `/bin/hostname -s` trafd.sh: (stop) warning: pid_filename not found"
	fi
    done
;;
*)
        /bin/echo "Usage: `/usr/bin/basename $0` {start|stop}" >&2
        exit 64
        ;;
esac

exit 0
