/*
 *	Copyright (c) 1999 RISS-Telecom Networking Center
 *
 *	Copyright (c) 1993 The CAD lab of the
 *	Novosibirsk Institute of Broadcasting and Telecommunication
 *
 *	BPFT $Id: traffic.h,v 1.5 2002/10/18 03:28:27 stas_degteff Exp $ Exp
 *
 *
 * Redistribution and use in source forms, with and without modification,
 * are permitted provided that this entire comment appears intact.
 * Redistribution in binary form may occur without any restrictions.
 *
 * THIS SOFTWARE IS PROVIDED ``AS IS'' WITHOUT ANY WARRANTIES OF ANY KIND.
 */

/*	File with major definitions	*/

#include <sys/time.h>
#include <syslog.h>
#include <netinet/in.h>

#ifndef OLD
#define OLD 0
#endif
#ifndef SRCDST
#define SRCDST 1
#endif
#ifndef NOPORT
#define NOPORT 2
#endif
#ifdef LAYOUT_OLD
#define LAYOUT OLD
#elif defined(LAYOUT_SRCDST)
#define LAYOUT SRCDST
#elif defined(LAYOUT_NOPORT)
#define LAYOUT NOPORT
#endif

#ifndef LAYOUT
/*#define LAYOUT OLD */      /* Old variant: only server-side port value */
#define LAYOUT SRCDST        /* Source & destination ports values present */
/*#define LAYOUT NOPORT */   /* Don't show port values. Calculate traffic for hosts only */
#endif

/*
 * Maximum size of the hashing memory table.
 * If you have big network traffic for example more then MAX_T_ENTRIES
 * records daily and you want backup it to file one per day then increase
 * it value.
 */
#define	MAX_T_ENTRIES		104800

/*
 * Maximum numder of the stored data records for autobackup to disk.
 */
#define MAX_TO_SAVE		MAX_T_ENTRIES - 75

/*
 * Minimum size of traffic for save record into file.
 * If 'all size' field value less, this record not saved.
 */
#ifndef MIN_SIZE_TO_SAVE
#define MIN_SIZE_TO_SAVE	1024L
#endif

/*
 * Magic number, this is offset used in hash algoritm to moved on
 * hash table. Do not edit, this is the nice numder.
 */
#define	E_OFFSET		7	/* XXX */

/*
 * This is the main hash function definition, produce start index
 * for begin hash table scan procedure. Very scientific formula. :)
 */
#define	HASHFUNC(p)	(((p)^((p)>>7)^((p)>>15)^((p)>>23)) % MAX_T_ENTRIES)

#define	EQUAL		0
#define	LESS		-1
#define	GREATER		1

/*
 * Single record's format description.
 */
#if LAYOUT==OLD
  struct t_entry {
	struct in_addr	in_ip, out_ip;	/* src ip addr and dst ip addr */
	u_char	ip_protocol;	/* which protocol been used (/etc/protocols) */
	u_char	who_srv;	/* who was server flag */
	u_short	p_port;		/* which port been used (/etc/services) */
	u_long	n_psize;	/* how many bytes in ip datagrams passed */
	u_long	n_bytes;	/* how many data bytes passed */
  };
#else
  /* branch 4.* */
  struct t_entry {
	struct in_addr  in_ip, out_ip;  /* src ip addr and dst ip addr */
	u_char  ip_protocol;    /* which protocol been used (/etc/protocols) */
	u_short o_port;		/* source port */
	u_short p_port;         /* destination port */
	u_long  n_psize;        /* how many bytes in ip datagrams passed */
	u_long  n_bytes;        /* how many data bytes passed */
  };
#endif

struct t_header {
	int    t_size;
	struct timeval start;
	struct timeval stop;
};

#ifndef	IPPROTO_OSPF
#define	IPPROTO_OSPF	89
#endif
#ifndef	IPPROTO_IGMP
#define	IPPROTO_IGMP	2
#endif

/*
 * Definition of the global counters.
 */
extern int n_entry;		/* counter for number of records */
extern struct t_entry *entries; /* collected traffic table */
