/*
 * Copyright (c) 2001 Stas Degteff <g@grumbler.org>
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that: (1) source code distributions
 * retain the above copyright notice and this paragraph in its entirety, (2)
 * distributions including binary code include the above copyright notice and
 * this paragraph in its entirety in the documentation or other materials
 * provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND WITHOUT ANY EXPRESS OR IMPLIED
 * WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
 *
 *
 * BPFT - bpfilter trafic collector
 *
 * $Id: output.c,v 1.2 2002/10/18 03:28:27 stas_degteff Exp $
 * Common functions for trafstat & traflog utilities.
 *
 */

#include <unistd.h>
#include <netinet/in.h>
#include <sys/param.h>
#include "output.h"
#include "traffic.h"

int
sortbyfrom(left, right)
	register struct t_entry *left;
	register struct t_entry *right;
{
	if (htonl(left->in_ip.s_addr) < htonl(right->in_ip.s_addr))
		return LESS;
	if (htonl(left->in_ip.s_addr) > htonl(right->in_ip.s_addr))
		return GREATER;
	return EQUAL;
}

int
sortbyto(left, right)
	register struct t_entry *left;
	register struct t_entry *right;
{
	if (htonl(left->out_ip.s_addr) < htonl(right->out_ip.s_addr))
		return LESS;
	if (htonl(left->out_ip.s_addr) > htonl(right->out_ip.s_addr))
		return GREATER;
	return EQUAL;
}

int
sortbypport(left, right)
	register struct t_entry *left;
	register struct t_entry *right;
{
	if (htons(left->p_port) < htons(right->p_port))
		return LESS;
	if (htons(left->p_port) > htons(right->p_port))
		return GREATER;
	return EQUAL;
}

#if LAYOUT!=OLD
int
sortbyoport(left, right)
        register struct t_entry *left;
        register struct t_entry *right;
{
        if (htons(left->o_port) < htons(right->o_port))
                return LESS;
        if (htons(left->o_port) > htons(right->o_port))
                return GREATER;
        return EQUAL;
}
#endif

int
sortbysize(left, right)
	register struct t_entry *left;
	register struct t_entry *right;
{
	if (left->n_psize < right->n_psize)
		return GREATER;
	if (left->n_psize > right->n_psize)
		return LESS;
	return EQUAL;
}


traf_print()
{
	register int i, j;
	u_int64_t abytes = 0, dbytes = 0;
	char buf[MAXHOSTNAMELEN + 1];
#if LAYOUT==OLD
	char *port, *user, *proto;
#else
	char *s_port, *d_port, *proto;
#endif
	for (i = 0; i < s.t_size; i++)
		abytes += sum[i].n_psize, dbytes += sum[i].n_bytes;
	if (rflag) {
		printf("%qu\n", abytes);
		return;
	}
	qsort(sum, s.t_size, sizeof(struct t_entry), sortbysize);
	switch ( order ){
	case ORDER_TO:
		qsort(sum, s.t_size, sizeof(struct t_entry), sortbyto);
		break;
	case ORDER_FROM:
		qsort(sum, s.t_size, sizeof(struct t_entry), sortbyfrom);
		break;		
	case ORDER_SIZE:
		qsort(sum, s.t_size, sizeof(struct t_entry), sortbysize);
		break;
	case ORDER_DPORT:
		qsort(sum, s.t_size, sizeof(struct t_entry), sortbypport);
		break;
#if LAYOUT!=OLD
	case ORDER_SPORT:
		qsort(sum, s.t_size, sizeof(struct t_entry), sortbyoport);
		break;
#endif
	}
	if (!fvnum) {
		gethostname(buf, MAXHOSTNAMELEN);
		printf("\n (%s) %s at", device_name, buf);
		printf(" %.15s -", ctime((time_t *)&s.start) + 4);
		printf(" %.15s\n", ctime((time_t *)&s.stop) + 4);
		printf(" Summary: %qu data bytes, %qu all bytes, %u records\n",
		       dbytes, abytes, s.t_size);
		printf("\
     From           Port         To            Port  Proto     Data       All\n");
	}
	for (i = 0; i < s.t_size; i++) {
#if LAYOUT==OLD
		port = "undef";
#else
		s_port = "";
		d_port = "";
#endif
		switch(sum[i].ip_protocol) {
			case IPPROTO_TCP:
				if (sum[i].p_port || sum[i].o_port)
#if LAYOUT==OLD
					port = tcpport_string(sum[i].p_port);
#else
					s_port = tcpport_string(sum[i].o_port);
					d_port = tcpport_string(sum[i].p_port);
#endif
				proto = "tcp";
				break;
			case IPPROTO_UDP:
				if (sum[i].p_port || sum[i].o_port)
#if LAYOUT==OLD
					port = udpport_string(sum[i].p_port);
#else
					s_port = udpport_string(sum[i].o_port);
					d_port = udpport_string(sum[i].p_port);
#endif
				proto = "udp";
				break;
			case IPPROTO_ICMP:
				proto = "icmp";
				break;
			case IPPROTO_EGP:
				proto = "egp";
				break;
			case IPPROTO_OSPF:
				proto = "ospf";
				break;
			case IPPROTO_IGMP:
				proto = "igmp";
				break;
			default:
				proto = "unkn";
		}
#if LAYOUT==OLD
		if (sum[i].p_port)	user = "client";
		else			user = "none";
#endif
		if (!fvnum) {
#if LAYOUT==OLD
			printf("%-18.18s %-6.6s  ", ipaddr_string(&sum[i].in_ip),
			       (sum[i].who_srv & 1) ? port : user);
			printf("%-18.18s %-6.6s  ", ipaddr_string(&sum[i].out_ip),
			       (sum[i].who_srv & 2) ? port : user);
			printf("%-4.4s %9ld %10ld\n", proto, sum[i].n_bytes,
			       sum[i].n_psize);
#else
			printf("%-18.18s %-6.6s  ", ipaddr_string(&sum[i].in_ip),
			       s_port);
			printf("%-18.18s %-6.6s  ", ipaddr_string(&sum[i].out_ip),
			       d_port);
			printf("%-4.4s %9ld %10ld\n", proto, sum[i].n_bytes,
			       sum[i].n_psize);
#endif
		} else 
			for (j = 0; j < fvnum; j++)
				switch (fv[j].cmd) {
				case FROM:
					printf(fv[j].format, ipaddr_string(&sum[i].in_ip));
					break;
				case TO:
					printf(fv[j].format, ipaddr_string(&sum[i].out_ip));
					break;
#if LAYOUT==OLD
				case SPORT:
					printf(fv[j].format,
					       (sum[i].who_srv & 1) ? port : user);
					break;
				case DPORT:
					printf(fv[j].format,
					       (sum[i].who_srv & 2) ? port : user);
					break;
#else
				case SPORT:
					printf(fv[j].format,
					       s_port);
					break;
				case DPORT:
					printf(fv[j].format,
					       d_port);
					break;
#endif
				case PROTO:
					printf(fv[j].format, proto);
					break;
				case BYTES:
					printf(fv[j].format, sum[i].n_bytes);
					break;
				case PSIZE:
					printf(fv[j].format, sum[i].n_psize);
					break;
				case FTIME:
					strftime(buf, sizeof(buf), fv[j].format,
						 localtime((time_t *)&s.start));
					printf("%s", buf);
					break;
				case LTIME:
					strftime(buf, sizeof(buf), fv[j].format,
						 localtime((time_t *)&s.stop));
					printf("%s", buf);
					break;
				}
	}
}


/*
 * Put traffic table to fd in binary form.
 */
void
traf_put(fd)
	register int fd;
{
	register int i;
	int c = 0;

	c += write(fd, (int *)&n_entry, sizeof(n_entry));
	c += write(fd, (struct timeval *)&start_time, sizeof(start_time));
	c += write(fd, (struct timeval *)&drop_time, sizeof(drop_time));
	for (i = 0; i < n_entry; i++) {
		c += write(fd, (struct t_entry *)&entries[i],
			   sizeof(struct t_entry));
	}
	(void)write(fd, (int *)&c, sizeof(c));
	return;
}

traf_write(fd)
	FILE *fd;
{
	if (fwrite((struct t_header *)&s, sizeof(struct t_header), 1, fd) != 1)
		error("write header error");
	if (fwrite((struct t_entry *)sum, sizeof(struct t_entry), s.t_size, fd) != s.t_size)
		error("write table error");
	return;
}
