/*
 *	Copyright (c) 1999 RISS-Telecom Networking Center
 *
 *	Copyright (c) 1993 The CAD lab of the
 *	Novosibirsk Institute of Broadcasting and Telecommunication
 *
 *	BPFT $Id: traffic.c,v 1.14 2004/04/14 11:44:13 stas_degteff Exp $
 *
 *
 * Redistribution and use in source forms, with and without modification,
 * are permitted provided that this entire comment appears intact.
 * Redistribution in binary form may occur without any restrictions.
 *
 * THIS SOFTWARE IS PROVIDED ``AS IS'' WITHOUT ANY WARRANTIES OF ANY KIND.
 */

/*	traffic.c - collect network traffic	*/

#include <stdio.h>
#include <string.h>
#include <unistd.h>
#include <errno.h>
#include <syslog.h>
#include <stdlib.h>
#ifdef USE_PIPE
#include <fcntl.h>
#include <signal.h>
#include <setjmp.h>
#else /* Unix Domain socket */
#include <sys/socket.h>
#include <sys/un.h>
#endif
#include <sys/types.h>
#include <sys/param.h>
#include <sys/time.h>
#include <sys/stat.h>
#include <netinet/in.h>
#include <netinet/in_systm.h>
#include <netinet/ip.h>
#include <netinet/udp.h>
#include <netinet/tcp.h>
#ifndef __linux__
#include <netinet/ip_var.h>
#include <netinet/udp_var.h>
#include <netinet/tcpip.h>
#endif

#include <netdb.h>
#include <pcap.h>

#include "interface.h"
#include "traffic.h"
#include "trafd.h"


#ifdef USE_PIPE
static jmp_buf after_pipe;
#endif

static int iplen;
static int dlen;
typedef struct t_entry *p_entry;
		/*static struct t_entry t;*/
static struct timeval begin_time;
#if LAYOUT==OLD
  /* This array collect port-proto from /etc/services.
     Used for detection http-proxy, irc, icq & etc. servers.
     Very non-precision method (map 64511 to 1024)
  */
  static u_char s_port_big[IPPORT_RESERVED];
#endif

/*
 * Initialize traffic collector.
 */
int
traf_init(restore)
	int restore;
{
	register int i;
#if LAYOUT==OLD
	struct servent *sv;
#endif
	
	i = MAX_T_ENTRIES * sizeof(struct t_entry);
	if ((entries = (struct t_entry *) valloc(i)) == NULL)
		return (i);
	(void)bzero((struct t_entry *)entries, i);
#if LAYOUT==OLD
	for(i = 0; i < IPPORT_RESERVED; i++)
		s_port_big[i] = 0;
	setservent(1);
	while (sv = getservent()) {
		NTOHS(sv->s_port);
		if (sv->s_port > IPPORT_RESERVED) {
			i = sv->s_port & (IPPORT_RESERVED-1);
			if (strcmp(sv->s_proto, "tcp") == 0)
				s_port_big[i] |= IPPROTO_TCP;
			else if (strcmp(sv->s_proto, "udp") == 0)
				s_port_big[i] |= IPPROTO_UDP;
		}
	}
	endservent();
#endif
	n_entry = 0;
	if (restore)
		traf_restore();
	else
		gettimeofday(&begin_time, (struct timezone *)NULL);
	return 0;
}

/*
 * Hash by in_ip.s_addr.
 */
static inline unsigned
hash( e )
	register p_entry e;
{
	return HASHFUNC(e->in_ip.s_addr);
}

int inline IsEntryFound(unsigned index, struct t_entry *PacketData) {
    return
	PacketData->in_ip.s_addr  == entries[index].in_ip.s_addr  &&
	PacketData->out_ip.s_addr == entries[index].out_ip.s_addr &&
	PacketData->ip_protocol   == entries[index].ip_protocol   &&
#if LAYOUT==OLD
	PacketData->who_srv       == entries[index].who_srv       &&
#else
	PacketData->o_port        == entries[index].o_port        &&
#endif
	PacketData->p_port        == entries[index].p_port;
}

int inline IsEmptyEntry(unsigned index) {
    return entries[index].in_ip.s_addr == NULL;
}

unsigned inline NextIndex(unsigned index) {
    return (index + E_OFFSET) % MAX_T_ENTRIES;
}

void InsertOrIncrementEntry(struct t_entry* PacketData) {
    unsigned packetHash = hash(PacketData);
    unsigned curIndex   = packetHash;
    char *Message = "(%s) table full, autosave";
    do { 
	if( IsEmptyEntry(curIndex) ) {
	    if( n_entry >= MAX_TO_SAVE )
		break;
	    entries[curIndex] = *PacketData;
	    n_entry ++;
	    return;
	}
	if( IsEntryFound(curIndex,PacketData) ) {
	    u_long
		new_n_psize = entries[curIndex].n_psize + PacketData->n_psize,
		new_n_bytes = entries[curIndex].n_bytes + PacketData->n_bytes;
	    if(
		new_n_psize < entries[curIndex].n_psize ||
		new_n_bytes < entries[curIndex].n_bytes
	    ) {
		Message = "(%s) counter overflowed, autosave";
		break;
	    }
	    entries[curIndex].n_psize = new_n_psize;
	    entries[curIndex].n_bytes = new_n_bytes;
	    return;
	}
	curIndex = NextIndex(curIndex);
    } while( curIndex != packetHash );
    syslog(LOG_WARNING, Message, device_name);
    ForkAndSave( curIndex == packetHash ? SAVEFORCECLEAR: SAVE );
    entries[packetHash] = *PacketData;
    n_entry = 1;
}

#if LAYOUT==OLD
void inline calcPortAndServer(
    u_short th_sport, u_short th_dport, int proto
    u_short *p_port, u_char *who_srvr
) {
    th_sport = ntohs(th_sport);
    th_dport = ntohs(th_dport);
    if( th_sport < th_dport ) {
	*p_port = th_sport;
	*who_srv = 1;
    }
    else if( th_sport > th_dport ) {
	*p_port = th_dport;
	*who_srv = 2;
    }
    else if( th_sport == th_dport ) {
	*p_port = th_sport;
	*who_srv = 3;
    }

    if (*p_port > IPPORT_RESERVED) {
	if( s_port_big[th_sport & (IPPORT_RESERVED-1)] & proto ) {
	    *p_port = th_sport;
	    *who_srv = 1;
	}
	else if( s_port_big[th_dport & (IPPORT_RESERVED-1)] & proto ) {
	    *p_port = th_dport;
	    *who_srv = 2;
	}
	if (th_sport == th_dport) *who_srv = 3;
    }
}
#endif

/*
 * Calculate data length and server port in tcp packets.
 */
void
traf_tcp(tp, ip)
	register struct tcphdr *tp;
	register struct ip *ip;
{
	struct t_entry t;

	if ((u_char *)(tp + 1) > snapend || dlen < sizeof(struct tcphdr))
		return;

	t.in_ip.s_addr = ip->ip_src.s_addr;
	t.out_ip.s_addr = ip->ip_dst.s_addr;
	t.ip_protocol = IPPROTO_TCP;
	t.n_psize = (u_long)iplen;
	t.n_bytes = (u_long)(dlen - tp->th_off * 4) & 2047L;

#if LAYOUT==OLD
	calcProtAndWhoServer(
	    tp->th_sport,tp->th_dport,IPPROTO_TCP,
	    &t.p_port, &t.who_srvr
	);
#else
	t.p_port = ntohs(tp->th_dport);
	t.o_port = ntohs(tp->th_sport);
#endif
	//traf_add(insertentry(&t), t.n_bytes, t.n_psize);
	InsertOrIncrementEntry(&t);
}

/*
 * Calculate data length and server port in udp packets.
 */
void
traf_udp(up, ip)
	register struct udphdr *up;
	register struct ip *ip;
{
	register u_char *cp = (u_char *)(up + 1);
	struct t_entry t;

	if (cp > snapend || dlen < sizeof(struct udphdr))
		return;

	t.in_ip.s_addr = ip->ip_src.s_addr;
	t.out_ip.s_addr = ip->ip_dst.s_addr;
	t.ip_protocol = IPPROTO_UDP;
	t.n_psize = (u_long)iplen;
	t.n_bytes = (u_long)(ntohs(up->uh_ulen) - sizeof(*up)) & 2047L;

#if LAYOUT==OLD
	calcProtAndWhoServer(
	    up->uh_sport,up->uh_dport,IPPROTO_UDP,
	    &t.p_port, &t.who_srvr
	);
#else
	t.o_port = ntohs(up->uh_sport);
	t.p_port = ntohs(up->uh_dport);
#endif
	//traf_add(insertentry(&t), t.n_bytes, t.n_psize);
	InsertOrIncrementEntry(&t);

	return;
}

/*
 * All other ip packets also collect.
 */
void
traf_ip(ip)
	register struct ip *ip;
{
	struct t_entry t;

	t.in_ip.s_addr = ip->ip_src.s_addr;
	t.out_ip.s_addr = ip->ip_dst.s_addr;
	t.ip_protocol = ip->ip_p;
	t.n_psize = (u_long)iplen;
	t.n_bytes = (u_long)dlen;
#if LAYOUT==OLD
	t.p_port = t.who_srv = 0;
#else
	t.p_port = t.o_port = 0;
#endif
	//traf_add(insertentry(&t), t.n_bytes, t.n_psize);
	InsertOrIncrementEntry(&t);
	return;
}

/*
 * All ip packets with zeroed ports & protocol fields collect.
 */
void
traf_ip_noproto(ip)
	register struct ip *ip;
{
	struct t_entry t;

	t.in_ip.s_addr = ip->ip_src.s_addr;
	t.out_ip.s_addr = ip->ip_dst.s_addr;
	t.ip_protocol = 0;
	t.n_psize = (u_long)iplen;
	t.n_bytes = (u_long)dlen;
#if LAYOUT==OLD
	t.p_port = t.who_srv = 0;
#else
	t.p_port = t.o_port = 0;
#endif
	//traf_add(insertentry(&t), t.n_bytes, t.n_psize);
	InsertOrIncrementEntry(&t);
	return;
}


void ForkAndSave(SaveTypeEnum SaveType) {
    int forkRetCode;
    char *file_name = (SaveType==SAVEBACKUP)?file_out:file_backup;
    char *save_mode = (SaveType==SAVEBACKUP)?"w":"a";

    if( (forkRetCode=fork()) == 0 ) {
	traf_save(file_name, save_mode);
	exit(0);
    }
    else if( forkRetCode < 0 ) {
	if( SaveType == SAVEFORCECLEAR ) {
	    syslog(
		LOG_CRIT,
		"(%s) can't start save process: %m, traffic table is overloaded and cleared",
		device_name
	    );
	    traf_clear();
	}
	else {
	    syslog(
		LOG_ERR,
		"(%s) can't start %s process: %m, traffic table isn't cleared",
		device_name,
		(SaveType==SAVEBACKUP?"backup":"save")
	    );
	}
    }
    else
	traf_clear();
}

/*
 * Drop data to disk.
 * If mode = "w" then dump full table, else drop only essential data.
 */
void
traf_save(fname, mode)
	char *fname;
	char *mode;
{
	register FILE *fd;
	register int i;
	struct timeval now;
	int c = 0;
	unsigned ess_skip=0, ess_entry = 0;
	char *msg;
	struct pcap_stat pcap_status={0,0,0};
	char  fname_tmp[PATH_MAX+4];

	/* If pcap descriptor is good, get status of the pcap and check
	 * to dropped packets.
	 * Condition expression don't beatiful, hoverer this work right.
	 * Slow?  It doesn't matter there! */

	if ((*mode == 'a') && ((fd = fopen(fname,"r")) != NULL)) {
		fclose(fd);
		if(rename(fname, file_tmp) != 0) {
			(void)syslog(LOG_ERR, "(%s) Can't rename %s to %s: %m",
				     device_name, fname, file_tmp);
			return;
		}
	}
	if ((fd = fopen(file_tmp, mode)) == NULL) {
		(void)syslog(LOG_ERR, "(%s) fopen %s: %m",
			     device_name, file_tmp);
		return;
	}
	gettimeofday(&now, (struct timezone *)NULL);
	if (*mode == 'w') {
		int max = MAX_T_ENTRIES;

		c += fwrite((int *)&max, sizeof(max), 1, fd);
		c += fwrite((int *)&n_entry, sizeof(n_entry), 1, fd);
		ess_entry = n_entry;
		msg = "dumped";
	} else {
		for (i = 0; i < MAX_T_ENTRIES; i++)
			if (entries[i].in_ip.s_addr != 0) {
			    if( entries[i].n_psize >= minsize )
				ess_entry++;
			    else ess_skip++;
			}
		c += fwrite((int *)&ess_entry, sizeof(ess_entry), 1, fd);
		msg = "append";
	}
	c += fwrite((struct timeval *)&begin_time, sizeof(begin_time), 1, fd);
	c += fwrite((struct timeval *)&now, sizeof(now), 1, fd);
	if (*mode == 'w') {
		c += fwrite((struct t_entry *)entries, sizeof(struct t_entry),
			    MAX_T_ENTRIES, fd);
		(void)fwrite((int *)&c, sizeof(c), 1, fd);
	} else
		for (i = 0; i < MAX_T_ENTRIES; i++) {
			if (entries[i].in_ip.s_addr == 0
			   || entries[i].n_psize < minsize
			)	continue;
			fwrite((struct t_entry *)&entries[i],
			       sizeof(struct t_entry), 1, fd);
		}
	(void)fclose(fd);

	if(rename(file_tmp, fname) !=0) {
		(void)syslog(LOG_ERR, "(%s) Can't rename %s to %s: %m",
			     device_name, file_tmp, fname);
		return;
	}

	if( pd!=NULL )
	{ if( pcap_stats(pd, &pcap_status) ){
		(void)syslog(LOG_ERR, "(%s) pcap_stats(): %s",
		     device_name, pcap_geterr(pd) );		
	  }else if( pcap_status.ps_drop ){
		if( ess_skip )
		   (void)syslog(LOG_WARNING, "(%s) %u packets received, "
			"%u packets dropped, %u records %s to %s, "
			"%u records less %lu (not saved)",
			device_name, pcap_status.ps_recv, pcap_status.ps_drop,
			ess_entry, msg, fname, ess_skip, minsize );
		else
		   (void)syslog(LOG_WARNING, "(%s) %u packets received, "
			"%u packets dropped, %u records %s to %s",
			device_name, pcap_status.ps_recv, pcap_status.ps_drop,
			ess_entry, msg, fname );

	  }else if( !pcap_status.ps_drop ){
		if( ess_skip )
		   (void)syslog(LOG_INFO, "(%s) %u packets received, "
			"%u records %s to %s, %u records less %lu (not saved)",
			device_name, pcap_status.ps_recv, ess_entry, msg, fname,
			ess_skip, minsize );
		else
		   (void)syslog(LOG_INFO, "(%s) %u packets received, "
			"%u records %s to %s",
			device_name, pcap_status.ps_recv, ess_entry, msg, fname );
	  }
	}
	return;
}

/*
 * Restore data from disk into memory. Useful when need to resume collect,
 * for example, after shutdown.
 */
void
traf_restore()
{
	register FILE *fd;
	register int c = 0;
	struct timeval now;
	int max, count;

	if ((fd = fopen(file_out, "r")) == NULL) {	/*ITS4: ignore*/
		gettimeofday(&begin_time, (struct timezone *)NULL);
		return;
	}
	c += fread((int *)&max, sizeof(max), 1, fd);
	c += fread((int *)&n_entry, sizeof(n_entry), 1, fd);
	c += fread((struct timeval *)&begin_time, sizeof(begin_time), 1, fd);
	c += fread((struct timeval *)&now, sizeof(now), 1, fd);
	c += fread((struct t_entry *)entries, sizeof(struct t_entry),
		   MAX_T_ENTRIES, fd);
	(void)fread((int *)&count, sizeof(count), 1, fd);
	(void)fclose(fd);
	unlink(file_out);	/*ITS4: ignore*/
	if (c != count || max != MAX_T_ENTRIES) {
		(void)syslog(LOG_WARNING, "(%s) can't restore from %s",
			     device_name, file_out);
		n_entry = MAX_T_ENTRIES;
		(void)traf_clear();
	} else
		(void)syslog(LOG_NOTICE, "(%s) %d records restored from %s",
			     device_name, n_entry, file_out);
}

/*
 * Full clear hash table, start time and misc counters.
 */
void
traf_clear()
{
	(void)bzero((struct t_entry *)entries,
		    sizeof(struct t_entry) * MAX_T_ENTRIES);
	gettimeofday(&begin_time, (struct timezone *)NULL);
	(void)syslog(LOG_INFO, "(%s) %d records cleared",
		     device_name, n_entry);
	n_entry = 0;
	return;
}

/*
 * Send data via ipc, some utility like trafstat may use it.
 */
void
traf_pipe()
{
	register int fd, i;
	struct timeval now;
	int c = 0, ess_entry =0;
#ifdef USE_PIPE
	void onpipe();
#else /* UNIX Domain socket */
	struct sockaddr_un server;
#endif
	/* Calculate number of records contained essential data */
	for (i = 0; i < MAX_T_ENTRIES; i++)
		if (entries[i].in_ip.s_addr != 0 &&
		    entries[i].n_psize >= 2048L)
			ess_entry++;

#ifdef	USE_PIPE
	/* Prepare reaction on SIGPIPE interrupt */
	if (setjmp(after_pipe)) {
		close(fd);
		(void)signal(SIGPIPE, SIG_IGN);
		return;
	}
	(void)signal(SIGPIPE, onpipe);

	/* Open named pipe */
	if ((fd = open(file_fifo, O_WRONLY /* |O_NONBLOCK */)) < 0) {
		(void)syslog(LOG_WARNING, "(%s) open %s: %m",
			     device_name, file_fifo);
		return;
	}
#else	/* Using UNIX IPC Domain */
	if ((fd = socket(AF_UNIX, SOCK_STREAM, 0)) < 0) {
		(void)syslog(LOG_WARNING, "(%s) open socket: %m", device_name);
		return;
	}
	server.sun_family = AF_UNIX;
        strncpy(server.sun_path, file_fifo, sizeof server.sun_path);

	if (connect(fd, (struct sockaddr *)&server,
		    sizeof(struct sockaddr_un)) < 0) {
		close(fd);
		(void)syslog(LOG_WARNING, "(%s) connect: %m", device_name);
		return;
	}
#endif /* USE_PIPE */
	gettimeofday(&now, (struct timezone *)NULL);
	c += write(fd, (int *)&ess_entry, sizeof(ess_entry));
	c += write(fd, (struct timeval *)&begin_time, sizeof(begin_time));
	c += write(fd, (struct timeval *)&now, sizeof(now));
	for (i = 0; i < MAX_T_ENTRIES; i++) {
		/*
		 * Don't send wholly memory table to ipc,
		 * only records contained essential data.
		 */
		if (entries[i].in_ip.s_addr == 0 ||
		    entries[i].n_psize < 2048L)
			continue;
		c += write(fd, (struct t_entry *)&entries[i],
			   sizeof(struct t_entry));
	}
	(void)write(fd, (int *)&c, sizeof(c));
	(void)close(fd);
#ifdef	USE_PIPE
	(void)syslog(LOG_INFO, "(%s) %d records sent to named pipe %s",
		     device_name, ess_entry, file_fifo);
	(void)signal(SIGPIPE, SIG_IGN);
#else
	(void)syslog(LOG_INFO, "(%s) %d records sent to socket %s",
		     device_name, ess_entry, file_fifo);
#endif
	return;
}

#ifdef	USE_PIPE
/*
 * Make clean return after SIGPIPE.
 */
void
onpipe()
{
	(void)syslog(LOG_WARNING, "(%s) broken pipe", device_name);
	longjmp(after_pipe, 1);
}
#endif

/*
 * This point will be called from bpf_readloop() per each packet.
 */
void
processing_ip(bp, length)
	register const u_char *bp;
	register uint length;
{
	register const struct ip *ip = (const struct ip*) bp;
	register int hlen;
	register unsigned char *cp;

	if ((u_char *)(ip + 1) > snapend)
		return;
	if (length < sizeof(struct ip)) {
		syslog(LOG_WARNING, "(%s) truncated-ip: discard %d bytes",
		       device_name, length);
		return;
	}
	iplen = (int)ntohs(ip->ip_len);
	if (length < iplen)
		syslog(LOG_INFO, "(%s) truncated-ip: %d bytes missing",
			device_name, iplen - length);
	hlen = ip->ip_hl * 4;
	dlen = iplen - hlen;
	if ( Xflag ){ /* Store full info or src&dst addresses only */
	/*
	 * If this fragment zero, hand it to the next higher level protocol,
	 * else it is fragmented datagram.
	 */
		if( !(ntohs(ip->ip_off) & 0x1fff)) {
			cp = (unsigned char *)ip + hlen;
			switch (ip->ip_p) {
			case IPPROTO_TCP:
				traf_tcp((struct tcphdr *)cp, ip);
				break;
			case IPPROTO_UDP:
				traf_udp((struct udphdr *)cp, ip);
				break;
			default:
				traf_ip(ip);
				break;
			}
		} else	traf_ip(ip);
	}else  traf_ip_noproto(ip);

	/* Check interrupt's queue */
	if (flag_hup) {
		alarm(0);
		flag_hup = 0;
		ForkAndSave(SAVEBACKUP);
	}
	if (flag_int) {
		alarm(0);
		flag_int = 0;
		ForkAndSave(SAVE);
	}
	if (flag_usr1) {
		int frk;
		flag_usr1 = 0;
		if (fork() == 0) {
			traf_pipe();
			exit(0);
		} else if (frk < 0) {
			(void)syslog(LOG_ERR, "(%s) can't start pipe process: %s",
			     device_name, strerror(errno));
		}
	}
	return;
}

/* The END */
