/*
 *
 *      Copyleft (c) 2003 Alexander Ilyin
 *
 * Redistribution and use in source forms, with and without modification,
 * are permitted provided that this entire comment appears intact.
 * Redistribution in binary form may occur without any restrictions.
 *
 * THIS SOFTWARE IS PROVIDED ``AS IS'' WITHOUT ANY WARRANTIES OF ANY KIND.
 */

/*  commondb.c - tcp/udp data traffic saver to the database, common part   */



#include <stdio.h>
#include <string.h>
#include "commondb.h"

#define PATH_TO_LOG "/var/opt/trafd/"
#define PATH_TO_SPOOL "/var/opt/trafd/spool/"
#define SPOOL_ENTRY_NAME "spoolentry"

/* TODO: use the config file instead following four defines */
#define DEV_NAME "eth0"

#define SQL_STRING "INSERT INTO traflog (RegTime,InIP,OutIP,TrafSize,DataSize) VALUES($1,$2,$3,$4,$5)"

#define SQL_PARAM_INFO "regStopTime:TIMESTAMP,inIP:INTEGER,outIP:INTEGER,ipDatagramsSize:BIGINT,bytesPassed:BIGINT"

#define DB_URL "host=localhost " \
               "dbname=traf " \
               "user=traf " \
               "password=Nbrhbn14"


void hfeRegStartTime(struct t_header* th, char* result); /*Extract regStartTime from t_header as TIMESTAMP*/
void hfeRegStopTime(struct t_header* th, char* result);  /*Extract regStopTime from t_header as TIMESTAMP*/

void efeInIp(struct t_entry* te, char* result);
void efeOutIp(struct t_entry* te, char* result);
void efeProtocol(struct t_entry* te, char* result);
void efeSourcePort(struct t_entry* te, char* result);
void efeDestPort(struct t_entry* te, char* result);
void efeIpDatagramsSize(struct t_entry* te, char* result);
void efeBytesPassed(struct t_entry* te, char* result);



char bufRegStartTime[TIMESTAMP_MAX_LENGTH];
char bufRegStopTime[TIMESTAMP_MAX_LENGTH];

char bufInIp[LONG_MAX_LENGTH];
char bufOutIp[LONG_MAX_LENGTH];
char bufProtocol[LONG_MAX_LENGTH];
char bufSourcePort[LONG_MAX_LENGTH];
char bufDestPort[LONG_MAX_LENGTH];
char bufIpDatagramsSize[LONG_MAX_LENGTH];
char bufBytesPassed[LONG_MAX_LENGTH];


/*Pay attanton to keep MAX_DATAITEMS_COUNT equal to count of this array*/
static struct t_dataitem possibleDataItems[] = {
                                  {"regStartTime", NULL, &hfeRegStartTime, NULL, bufRegStartTime},
                                  {"regStopTime", NULL, &hfeRegStopTime, NULL, bufRegStopTime},
                                  {"inIP", NULL, NULL, &efeInIp, bufInIp},
                                  {"outIP", NULL, NULL, &efeOutIp, bufOutIp},
                                  {"protocol", NULL, NULL, &efeProtocol, bufProtocol},
                                  {"sourcePort", NULL, NULL, &efeSourcePort, bufSourcePort},
                                  {"destPort", NULL, NULL, &efeDestPort, bufDestPort},
                                  {"ipDatagramsSize", NULL, NULL, &efeIpDatagramsSize, bufIpDatagramsSize},
                                  {"bytesPassed", NULL, NULL, &efeBytesPassed, bufBytesPassed}
                                 };



struct t_dataitem* currentDataItems[MAX_DATAITEMS_COUNT];
int dataItemsCount;



char *program_name;

char* pathToLog;
char* pathToSpool;
char* spoolEntryName;
char* sqlString;
char* sqlParamInfo;
char* dbUrl;
char* device_name;


char* device_name;

/* struct t_header defined in the traffic.h */
static struct t_header s;


int findDataEntry(char* name) {
    int i;
    for(i = 0; i < MAX_DATAITEMS_COUNT; i++) {
printf("Comparing %s WITH %s\n", name, possibleDataItems[i].name);
        if(!strcmp(name, possibleDataItems[i].name)) {
            return i;
        }
    }
    return -1; /* not found */
}

int initParamInfo(char* sqlParmInfo) {
  char * pch;
  int curDataItemIdx;
  char sql_params_string[MAX_DATAITEM_TEXT_LENGTH*MAX_DATAITEMS_COUNT];

  strncpy(sql_params_string, sqlParmInfo, MAX_DATAITEM_TEXT_LENGTH*MAX_DATAITEMS_COUNT);
  dataItemsCount = 0;

  pch = strtok(sql_params_string, ":,");
  while ((pch != NULL) && (curDataItemIdx < MAX_DATAITEMS_COUNT)) {
printf("Found value=%s\n", pch);

    curDataItemIdx = findDataEntry(pch);
    if(curDataItemIdx > 0) {

        currentDataItems[dataItemsCount] = &possibleDataItems[curDataItemIdx];

        /*read the datatype*/
        pch = strtok (NULL, ":,");
        if(pch != NULL) {
printf("Found sql type=%s\n", pch);
            currentDataItems[dataItemsCount]->sqlType = pch;
            pch = strtok (NULL, ":,");
            dataItemsCount++;
        } else {
           error("You miss the sql type of paramener %s",  currentDataItems[dataItemsCount]->name);
           return -1;
        }
    } else {
        error("Wrong name of paramener %s", pch);
        return -1;
    }
  }
  return 0;
}


void initParams() {

  pathToLog = PATH_TO_LOG;
  pathToSpool = PATH_TO_SPOOL;
  spoolEntryName = SPOOL_ENTRY_NAME;
  sqlString = SQL_STRING;
  dbUrl = DB_URL;
  device_name = DEV_NAME;

  initParamInfo(SQL_PARAM_INFO);
}

/* Get header from trafd.log */
int
getheader(register FILE *fd, register struct t_header *header)
{
  if (fread((struct t_header *)header, sizeof(struct t_header), 1, fd) != 1) {
    if (!feof(fd)) {
      warning("fread: can't read header (error: %s", strerror(errno) );
    }
    return 0;
  }
  if (header->t_size > MAX_TO_SAVE) {
    warning("table too big to fit into memory");
    return 0;
  }
  return 1;
}


int moveTheCurentLog() {
  char spoolEntryName[PATH_MAX];
  int fd;
  char buf[PATH_MAX];

  makeTheSpoolEntryName(spoolEntryName, device_name, ".XXXXXX");
  //printf("%s %d\n", spoolEntryName, PATH_MAX);

  //printf("Moving the current traflog to spool...\n");
  if((fd = mkstemp(spoolEntryName)) < 0){
          perror("mkstemp failed");
          return 1;
  }

  close(fd);
  //printf("Spool name: %s\n", spoolEntryName);
  snprintf(buf, sizeof(buf), "%strafd.%s", PATH_TO_LOG, device_name);

  if(rename(buf, spoolEntryName) != 0) {
        perror("rename failed");
          return 1;
  }
  return 0;
}


void makeTheSpoolEntryName(char* buf, char* device_name, char* suffix) {
    snprintf(buf, PATH_MAX, "%s%s.%s%s", pathToSpool, spoolEntryName, device_name, suffix);
}


int  processSpoolEntryFile(char* spoolEntryFileName, void* connInfo) {
    int retval = 0;
    register FILE *fd;
    char fullFileName[PATH_MAX];

    snprintf(fullFileName, sizeof(fullFileName), "%s%s", pathToSpool, spoolEntryFileName);

    if ((fd = fopen(fullFileName, "r")) == NULL) {
        error("can't open %s", fullFileName);
        return 1;
    }

    retval = processLogFile(fd, connInfo);
    fclose(fd);
    if(remove(fullFileName) != 0) {
        error("can't remove after processing %s", fullFileName);
        retval = 1;
    }

    return retval;
}


int walkThroughTheDir(void* connInfo, int errorCode) {
    if(errorCode == 0) {
        DIR *dirp;
        struct dirent *dirEntry;
        char spoolEntryNamePattern[PATH_MAX];

        snprintf(spoolEntryNamePattern,  sizeof(spoolEntryNamePattern), "%s.%s", spoolEntryName, device_name);
        int spoolEntryNamePatternLength = strnlen(spoolEntryNamePattern, PATH_MAX - 1);


        dirp = opendir(pathToSpool);
        while((dirEntry = readdir(dirp)) != NULL) {
            if(strncmp(dirEntry->d_name, spoolEntryNamePattern, spoolEntryNamePatternLength) == 0) {

                errorCode = processSpoolEntryFile(dirEntry->d_name, connInfo);
                if(errorCode != 0) {
                    break;
                }
            }
        }

        closedir(dirp);
    }
    return errorCode;
}


int processLogFile(register FILE *fd, void* connInfo) {
    struct t_header h;
    struct t_entry entry;
    int retval, i, j;
    int errors = 0;
    int headerParams[MAX_DATAITEMS_COUNT];
    int headerParamsCount = 0;
    int entryParams[MAX_DATAITEMS_COUNT];
    int entryParamsCount = 0;
    struct t_dataitem* curDataItem;

    //Prepare for some optimizations with retrieving the params values
    for(i = 0; i < dataItemsCount; i++) {
        if(currentDataItems[i]->headerFieldExtracter != NULL) {
            headerParams[headerParamsCount] = i;
            headerParamsCount++;
        }
        if(currentDataItems[i]->entryFieldExtracter != NULL) {
            entryParams[entryParamsCount] = i;
            entryParamsCount++;
        }
    }

    //Main loop through the headers in the log file
    while (getheader(fd, &h)) {
        //init the global header params
        for(j = 0; j < headerParamsCount; j++) {
            curDataItem = currentDataItems[headerParams[j]];
            curDataItem->headerFieldExtracter(&h, curDataItem->resultBuf);
        }

        //walk through the entries of the current header and save it to the DB
        for (i = 0; i < h.t_size; i++) {
            if (fread(&entry, sizeof(struct t_entry), 1, fd) != 1) {
                warning("processLogFile(): fread: can't read table");
                return 1;
            }

            //init entry params
            for(j = 0; j < entryParamsCount; j++) {
                curDataItem = currentDataItems[entryParams[j]];
                curDataItem->entryFieldExtracter(&entry, curDataItem->resultBuf);
            }

            saveToDatabase(connInfo);
        }
    }
    return 0;
}


void hfeRegStartTime(struct t_header* th, char* result) {
    struct tm bdt;
    localtime_r((time_t *)&th->start, &bdt);
    snprintf(result, TIMESTAMP_MAX_LENGTH,"%d-%.2d-%.2d %.2d:%.2d:%.2d",
                                        bdt.tm_year + 1900,
                                        bdt.tm_mon + 1,
                                        bdt.tm_mday,
                                        bdt.tm_hour,
                                        bdt.tm_min,
                                        bdt.tm_sec);
}

void hfeRegStopTime(struct t_header* th, char* result) {
    struct tm bdt;
    localtime_r((time_t *)&th->stop, &bdt);
    snprintf(result, TIMESTAMP_MAX_LENGTH,"%d-%.2d-%.2d %.2d:%.2d:%.2d",
                                        bdt.tm_year + 1900,
                                        bdt.tm_mon + 1,
                                        bdt.tm_mday,
                                        bdt.tm_hour,
                                        bdt.tm_min,
                                        bdt.tm_sec);
}

void efeInIp(struct t_entry* te, char* result) {
    snprintf(result, LONG_MAX_LENGTH, "%ld", ntohl(te->in_ip.s_addr));
}

void efeOutIp(struct t_entry* te, char* result) {
    snprintf(result, LONG_MAX_LENGTH, "%ld", ntohl(te->out_ip.s_addr));
}

void efeProtocol(struct t_entry* te, char* result) {
    snprintf(result, LONG_MAX_LENGTH, "%ld", te->ip_protocol);
}

void efeSourcePort(struct t_entry* te, char* result) {
    snprintf(result, LONG_MAX_LENGTH, "%ld", te->o_port);
}

void efeDestPort(struct t_entry* te, char* result) {
    snprintf(result, LONG_MAX_LENGTH, "%ld", te->p_port);
}

void efeIpDatagramsSize(struct t_entry* te, char* result) {
    snprintf(result, LONG_MAX_LENGTH, "%lu", te->n_psize);
}

void efeBytesPassed(struct t_entry* te, char* result) {
    snprintf(result, LONG_MAX_LENGTH, "%lu", te->n_bytes);
}



