/*
 *
 *      Copyleft (c) 2003 Alexander Ilyin
 *
 * Redistribution and use in source forms, with and without modification,
 * are permitted provided that this entire comment appears intact.
 * Redistribution in binary form may occur without any restrictions.
 *
 * THIS SOFTWARE IS PROVIDED ``AS IS'' WITHOUT ANY WARRANTIES OF ANY KIND.
 */

/*  commondb.h - tcp/udp data traffic saver to the database, common part   */


#include <stdio.h>
#include <stdlib.h>
#include <errno.h>
#include <time.h>
#include <dirent.h>
#include <netinet/in.h>

#include "../include/traffic.h"
#include "../include/pathnames.h"


#define BUF_SIZE 256
#define LONG_MAX_LENGTH 25
#define TIMESTAMP_MAX_LENGTH 32

#define MAX_DATAITEMS_COUNT 9
#define MAX_DATAITEM_TEXT_LENGTH 40


//TODO 1. check function for validating init file

//TODO 2. check for MAX_DATAITEMS_COUNT overflow in param string and param duplications

/**
 There is list of avalaible params for insert DDL

 regStartTime - stop time of the interval (t_header.start)
 regStopTime - stop time of the interval (t_header.stop)

 inIP - source IP address as integer using ntohl() to correct byte order
 outIP - destination IP address as integer using ntohl() to correct byte order
 protocol - integer number of protocol
 sourcePort - integer source port
 destPort  - integer destination port
 ipDatagramsSize - bigint how many bytes in ip datagrams passed
 bytesPassed - bigint how many data bytes passed

 Feel free to ask me if you need some additional data or data in other format
(f.g. I can add Ip in string format)

What params you want to be used can be setted in the ini file
params=<paramName1>:<SQL type of param1> <paramName2>:<SQL type of param2> etc

Example:
params=regStopTime:TIMESTAMP inIP:INTEGER outIP:INTEGER bytesPassed:BIGINT


*/

struct t_dataitem {
  char* name;
  char* sqlType;
  void (*headerFieldExtracter) (struct t_header* th, char* result);
  void (*entryFieldExtracter) (struct t_entry* te, char* result);
  char* resultBuf;
};


extern struct t_dataitem* currentDataItems[];
extern int dataItemsCount;

extern char *program_name;

extern char* pathToLog;
extern char* pathToSpool;
extern char* spoolEntryName;
extern char* sqlString;
extern char* sqlParamInfo;

extern char* dbUrl;

extern char* device_name;


void initParams();
int getheader(register FILE *fd, register struct t_header *header);
int moveTheCurentLog();
void makeTheSpoolEntryName(char* buf, char* device_name, char* suffix);
int walkThroughTheDir(void* connInfo, int errorCode);

/* Just these 2 functions is DB specific and its should be implemented for each used DB */
int processTheSpool();
void saveToDatabase(void* connInfo);

