/*
 *
 *      Copyleft (c) 2003 Alexander Ilyin
 *
 * Redistribution and use in source forms, with and without modification,
 * are permitted provided that this entire comment appears intact.
 * Redistribution in binary form may occur without any restrictions.
 *
 * THIS SOFTWARE IS PROVIDED ``AS IS'' WITHOUT ANY WARRANTIES OF ANY KIND.
 */

/*  pgsql.c - PostgreSQL specific part   */

#include "commondb.h"
#include <libpq-fe.h>


#define PREPARED_STM_NAME "tle1"
#define MAX_PREPARED_SMT_LENGTH 250
#define MAX_SQLDATATYPE_LENGTH 15


const char* paramValues[MAX_DATAITEMS_COUNT];

void prepareStatement(char* sqlString, char* result) {
    char dataTypes[MAX_DATAITEMS_COUNT*MAX_SQLDATATYPE_LENGTH];
    int i;

    dataTypes[0] = '\0';

    for(i = 0; i < dataItemsCount; i++) {
        if(i > 0) {
            strcat(dataTypes, ",");
        }
        strncat(dataTypes, currentDataItems[i]->sqlType, MAX_SQLDATATYPE_LENGTH);

        paramValues[i] = currentDataItems[i]->resultBuf;

    }
    sprintf(result, "PREPARE %s (%s) AS %s", PREPARED_STM_NAME, dataTypes, sqlString);
}

/*Prepare common db staff for handling data and call the handling itself*/
int processTheSpool() {
        int retval = 0;
        PGconn* conn;
        PGresult* res;
        char preparedStm[MAX_PREPARED_SMT_LENGTH];

        prepareStatement(sqlString, preparedStm);

        /* try to connect (synchronous) */
        conn = PQconnectdb(dbUrl);
        /* check connection status */
        if (PQstatus(conn) != CONNECTION_OK) {
                /* error during connection */
                fprintf(stderr, "Connection failed: %s\n", PQerrorMessage(conn));
                return 1;
        }


        res = PQexec(conn, preparedStm);

        /* compare result with PGRES_COMMAND_OK! (not PGRES_TUPLES_OK) */
        if (PQresultStatus(res) != PGRES_COMMAND_OK) {
                fprintf(stderr, PQresultErrorMessage(res));
                retval = 1;
        }
        PQclear(res);

        retval = walkThroughTheDir(conn, retval);

        PQfinish(conn);

        return retval;
}

void saveToDatabase(void* connInfo) {
    PGconn* conn = (PGconn*) connInfo;
    PGresult* res = PQexecPrepared(conn,
                                   PREPARED_STM_NAME,
				   dataItemsCount,
                                   paramValues,
                                   NULL, /*paramLengths,*/
                                   NULL, /*paramFormats,*/
                                   1);

    /* compare result with PGRES_COMMAND_OK! (not PGRES_TUPLES_OK) */
    if (PQresultStatus(res) != PGRES_COMMAND_OK) {
        warning("During traf entry insertion: %s", PQresultErrorMessage(res));
    }
    PQclear(res);
}

